"""Core schema module."""

__version__ = "0.2.0"  # denote a pre-release for 0.1.0 with 0.1a1

from . import id
from ._core import (  # noqa
    dobject,
    dtransform,
    dtransform_in,
    dtransform_out,
    jupynb,
    lndb_schema_core,
    track_do,
    track_do_type,
    user,
)



from datetime import datetime as datetime
from enum import Enum
from typing import Optional, Union

from sqlmodel import Field, ForeignKeyConstraint, SQLModel, UniqueConstraint

from .id import id_dobject, id_dtransform, id_track


def utcnow():
    return datetime.utcnow().replace(microsecond=0)


class lndb_schema_core(SQLModel, table=True):  # type: ignore
    """Schema module version."""

    id: Optional[str] = Field(primary_key=True)
    user_id: str = Field(foreign_key="user.id")
    time_created: datetime = Field(default_factory=utcnow, nullable=False)
    time_updated: datetime = Field(default_factory=utcnow, nullable=False)


class user(SQLModel, table=True):  # type: ignore
    """User operating `lamindb`."""

    __table_args__ = (UniqueConstraint("email"),)
    id: Optional[str] = Field(primary_key=True)
    email: str
    time_created: datetime = Field(default_factory=utcnow, nullable=False)
    time_updated: datetime = Field(default_factory=utcnow, nullable=False)


class dobject(SQLModel, table=True):  # type: ignore
    """Data objects in storage & memory.

    Data objects often correspond to files.

    Storage ⟷ memory examples:
    - `.csv`, `.tsv`, `.feather`, `.parquet` ⟷ `pd.DataFrame`
    - `.h5ad`, `.h5mu`, or their zarr versions ⟷ `anndata.AnnData`, `mudata.MuData`
    - `.jpg`, `.png` ⟷ `np.ndarray`, or a dedicated imaging in-memory container
    - zarr directory ⟷ zarr loader
    - TileDB store ⟷ TileDB loader
    - fastq ⟷ ?
    - .vcf ⟷ ?
    """

    id: Optional[str] = Field(default_factory=id_dobject, primary_key=True)
    v: str = Field(default=None, primary_key=True)
    name: Optional[str]
    file_suffix: str
    dsource_id: str = Field(foreign_key="dtransform.id")
    time_created: datetime = Field(default_factory=utcnow, nullable=False)
    time_updated: datetime = Field(default_factory=utcnow, nullable=False)


class dtransform(SQLModel, table=True):  # type: ignore
    """Data transformation."""

    __table_args__ = (
        ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"],
            ["jupynb.id", "jupynb.v"],
            name="dtransform_jupynb",
        ),
        ForeignKeyConstraint(
            ["pipeline_id", "pipeline_v"],
            ["pipeline.id", "pipeline.v"],
            name="dtransform_pipeline",
        ),
    )
    id: str = Field(default_factory=id_dtransform, primary_key=True)
    jupynb_id: Union[str, None] = None
    jupynb_v: Union[str, None] = None
    pipeline_id: Union[str, None] = None
    pipeline_v: Union[str, None] = None


class dtransform_in(SQLModel, table=True):  # type: ignore
    __table_args__ = (
        ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="dtransform_in_dobject",
        ),
    )
    dtransform_id: str = Field(foreign_key="dtransform.id", primary_key=True)
    dobject_id: str = Field(primary_key=True)
    dobject_v: str = Field(primary_key=True)


class dtransform_out(SQLModel, table=True):  # type: ignore
    __table_args__ = (
        ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="dtransform_out_dobject",
        ),
    )
    dtransform_id: str = Field(foreign_key="dtransform.id", primary_key=True)
    dobject_id: str = Field(primary_key=True)
    dobject_v: str = Field(primary_key=True)


class jupynb(SQLModel, table=True):  # type: ignore
    """Jupyter notebook from which users operate lamindb."""

    id: str = Field(default=None, primary_key=True)
    v: str = Field(default=None, primary_key=True)
    name: Optional[str]
    user_id: str = Field(foreign_key="user.id")
    time_created: datetime = Field(default_factory=utcnow, nullable=False)
    time_updated: datetime = Field(default_factory=utcnow, nullable=False)


class pipeline(SQLModel, table=True):  # type: ignore
    """Pipeline interfacing lamindb."""

    id: str = Field(default=None, primary_key=True)
    v: str = Field(default=None, primary_key=True)


# ----------
# Access log
# ----------


class track_do_type(str, Enum):
    """Data access types."""

    ingest = "ingest"
    query = "query"
    update = "update"
    delete = "delete"
    load = "load"


class track_do(SQLModel, table=True):  # type: ignore
    """Data access log: do operations on the database."""

    __table_args__ = (
        ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"],
            ["jupynb.id", "jupynb.v"],
            name="track_do_jupynb",
        ),
        ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="track_do_dobject",
        ),
    )

    id: Optional[str] = Field(default_factory=id_track, primary_key=True)
    type: track_do_type = Field(nullable=False)  #: Data access type.
    user_id: str = Field(foreign_key="user.id", nullable=False)
    jupynb_id: str
    jupynb_v: str
    time: datetime = Field(default_factory=utcnow, nullable=False)
    dobject_id: str
    dobject_v: str



import random
import string


def id(n_char: int) -> str:
    base62 = string.digits + string.ascii_letters.swapcase()
    id = "".join(random.choice(base62) for i in range(n_char))
    return id


def id_dobject() -> str:
    """IDs for dobject.

    21 characters (62**21=4e+37 possibilities) outperform UUID (2*122=5e+36).
    """
    return id(n_char=21)


def id_dtransform() -> str:
    return id(n_char=22)


def id_user() -> str:
    """User ID with 8 base62 characters.

    Consistent with 1M users producing 1k notebooks.
    Safe for 100k users producing 10k notebooks.

    Allows >2e14 users.

    This is one of 2 IDs that are centralized.

    Collision probability in decentralized system is:

    ======= ===========
    n_users p_collision
    ======= ===========
    100k    2e-05
    1M      2e-03
    """
    return id(n_char=8)


def id_track() -> str:
    return id(n_char=24)


def id_secret() -> str:
    return id(n_char=40)


def id_instance() -> str:
    return id(n_char=12)



from lndb_schema_core import id


def test_id():
    assert len(id.id_dobject()) == 21
    assert len(id.id_user()) == 8
    assert len(id.id_secret()) == 40
    assert len(id.id_track()) == 24
    assert len(id.id_instance()) == 12



"""Rename table interface to jupynb.

Revision ID: 0560ee3d73dc
Revises:
Create Date: 2022-07-21 22:25:40.022247

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "0560ee3d73dc"
down_revision = None
branch_labels = None
depends_on = None


# the whole migration strategy relies on copy so heavily
# because SQLite doesn't support altering constraints


def upgrade() -> None:
    # rename interface to jupynb
    op.create_table(
        "jupynb",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.Column("time_updated", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.execute("insert into jupynb select * from interface")
    op.drop_table("interface")

    # fix references in dobject
    op.create_table(
        "dobject_",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("file_suffix", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("jupynb_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("jupynb_v", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.Column("time_updated", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"], ["jupynb.id", "jupynb.v"], name="dobject_jupynb"
        ),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.execute("insert into dobject_ select * from dobject")
    op.drop_table("dobject")
    op.rename_table("dobject_", "dobject")

    # fix references in track_do
    op.create_table(
        "track_do_",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("jupynb_id", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("jupynb_v", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("time", sa.DateTime(), nullable=False),
        sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("dobject_v", sqlmodel.sql.sqltypes.AutoString()),
        sa.ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="track_do_dobject",
        ),
        sa.ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"], ["jupynb.id", "jupynb.v"], name="track_do_jupynb"
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute(
        "insert into track_do_ (id, type, user_id, jupynb_id, time, dobject_id) "
        "select id, type, user_id, interface_id, time, dobject_id "
        "from track_do"
    )
    op.execute("update track_do_ SET jupynb_v = '1'")
    op.execute("update track_do_ SET dobject_v = '1'")
    op.drop_table("track_do")
    op.rename_table("track_do_", "track_do")


def downgrade() -> None:
    pass
